-- handles mario's actions and interactions and includes action utilities/constants

-- combines allocate_mario_action and hook_mario_action
function allocate_and_hook_mario_action(func, flags, interaction)
    local action = allocate_mario_action(flags)
    hook_mario_action(action, func, interaction or 0)
    return action
end

-- accelerate towards this speed. Caps by throwing out sub-subpixel component and lower
function smb_horizontal_acceleration(m, accel, approachSpeed)
    if m.vel.x == approachSpeed then return floor_or_ceil(approachSpeed) end

    -- double acceleration if our speed is in the opposite direction we are facing
    if m.vel.x ~= 0 and (m.faceAngle.y > 0) ~= (m.vel.x > 0) then
        accel = accel * 2
    end

    local newVelX = 0
    if m.vel.x > approachSpeed then
        newVelX = m.vel.x - accel
        if newVelX < approachSpeed then
            if approachSpeed == 0 then return 0 end
            newVelX = floor_or_ceil(m.vel.x)
        end
    else
        newVelX = m.vel.x + accel
        if newVelX > approachSpeed then
            if approachSpeed == 0 then return 0 end
            newVelX = floor_or_ceil(m.vel.x)
        end
    end

    return newVelX
end

-- accelerate towards the terminal velocity. Cap by throwing out subpixel component and lower
function smb_vertical_acceleration(m, gravity)
    local smb = gSMBStates[m.playerIndex]
    local terminalVel = TERMINAL_VEL
    if smb.inWhirlpool then -- behave differently in whirlpools
        terminalVel = TERMINAL_VEL_WHIRLPOOL
        gravity = gravity + 1
    end
    if m.vel.y == -terminalVel then return truncate(-terminalVel, PX) end

    local newVelY = m.vel.y - gravity
    if newVelY <= -terminalVel then
        newVelY = truncate(-terminalVel, PX)
    end
    return newVelY
end

-- move camera based on position (should only run for local player)
prevMPosForScroll = 0
function smb_move_camera(m, allowBackwards)
    local smb = gSMBStates[m.playerIndex]
    local mPosForScroll = (m.pos.x * SSPX - 8) - smb.cameraX
    if prevMPosForScroll < mPosForScroll and mPosForScroll >= 0x50 then
        if mPosForScroll >= 0x70 then
            smb.cameraX = smb.cameraX + (mPosForScroll - prevMPosForScroll)
        else
            smb.cameraX = smb.cameraX + (mPosForScroll - prevMPosForScroll) // 1.5
        end
    elseif allowBackwards and prevMPosForScroll > mPosForScroll and mPosForScroll <= 0x70 then
        if mPosForScroll <= 0x50 then
            smb.cameraX = smb.cameraX + (mPosForScroll - prevMPosForScroll)
        else
            smb.cameraX = smb.cameraX + math.ceil((mPosForScroll - prevMPosForScroll) / 1.5)
        end
    end

    local activeScrollStop = (loaded_tile_data.scrollStop or 0) * BLOCK
    if smb.tempScrollStop ~= 0 then
        activeScrollStop = smb.tempScrollStop
    end

    if smb.cameraX < 0 then
        smb.cameraX = 0
    elseif activeScrollStop ~= 0 and smb.cameraX > (activeScrollStop // PX) + PX - NES_WIDTH then
        smb.cameraX = (activeScrollStop // PX) + PX - NES_WIDTH
    end
    prevMPosForScroll = (m.pos.x * SSPX - 8) - smb.cameraX
end

-- moves mario and checks collision
function smb_mario_step(m, skipCamera)
    local smb = gSMBStates[m.playerIndex]
    m.pos.x = m.pos.x + floor_or_ceil(m.vel.x)
    if (not smb.inWhirlpool) or m.vel.y >= 0 then
        m.pos.y = m.pos.y + floor_or_ceil(m.vel.y)
    else -- alt whirlpool gravity
        m.pos.y = m.pos.y + floor_or_ceil(2 * m.vel.y - smb.gravity)
    end

    -- move towards whirlpool
    if smb.inWhirlpool and m.usedObj and m.usedObj ~= m.marioObj then
        local o = m.usedObj
        if o.oPosX >= m.pos.x then
            m.pos.x = m.pos.x + 8
        else
            m.pos.x = m.pos.x - 8
        end
    end

    -- don't go past sides of the screen
    if m.pos.x - 8 * PX < smb.cameraX * PX then
        m.pos.x = (smb.cameraX + 8) * PX
        m.vel.x = m.vel.x % 1
    elseif m.pos.x + 8 * PX > (smb.cameraX + NES_WIDTH) * PX then
        m.pos.x = (smb.cameraX + NES_WIDTH - 8) * PX
        m.vel.x = m.vel.x % 1
    end

    -- move camera
    if m.playerIndex == 0 and not skipCamera then
        local allowBackwards = gGlobalSyncTable.allowBackwards and ((not warpZoneActive) or m.pos.y >= NES_HEIGHT * PX - 3 * BLOCK)
        smb_move_camera(m, allowBackwards)
    end

    -- check collision with multiple collision points
    local landed = false
    local hitWall = false
    local collider_offsets = get_collider_offsets(m)

    -- head check
    local topPoint = (NES_HEIGHT - 16) * PX
    if smb.isBig and not smb.crouch then
        topPoint = topPoint - 16 * PX
    end
    if m.pos.y + 2 * BLOCK <= topPoint then
        local tile, px = get_tile_at_mario_collider(m, collider_offsets[COLLIDER_HEAD])
        if tile ~= 0 then
            if coin_check(m, COLLIDER_HEAD, tile) then
                return false
            elseif m.vel.y > 0 and px < 11 then -- below top 5 pixels
                if is_tile_block(tile) then
                    if m.playerIndex == 0 then
                        play_smb_sound_stream(m.pos, "bump")
                        local blockX, blockY = get_block_coords_for_mario_collider(m, collider_offsets[COLLIDER_HEAD])
                        network_send_to_all_in_act(true, {
                            id = PACKET_BLOCK_INTERACT,
                            x = blockX,
                            y = blockY,
                            tile = tile,
                            coinBrickTimer = coin_brick_timer[blockX],
                            hitFromBelow = true,
                            big = smb.isBig,
                            from = network_global_index_from_local(0),
                            area = gNetworkPlayers[0].currAreaIndex,
                        })
                    end

                    if tile == TILE_BRICK then
                        m.vel.y = -0x20
                    else
                        m.vel.y = 0
                    end
                elseif is_tile_solid(tile) then
                    m.vel.y = -(m.vel.y % PX) - 0x10 -- flip subpixel and lower and set pixel to -1
                    play_smb_sound_stream(m.pos, "bump")
                    -- djui_chat_message_create(tostring(m.pos.y % BLOCK // PX))
                end
            end
        end
    end

    -- feet check
    local usedCollider = COLLIDER_LEFT_FOOT
    tile, px = get_tile_at_mario_collider(m, collider_offsets[usedCollider])
    local rightCollision = false
    if tile == 0 then
        rightCollision = true
        usedCollider = COLLIDER_RIGHT_FOOT
        tile, px = get_tile_at_mario_collider(m, collider_offsets[usedCollider])
    end
    
    if coin_check(m, usedCollider, tile) then return false end
    
    if tile ~= 0 and is_tile_solid(tile) and m.vel.y < 0 then
        if px >= 10 then -- above top 5 pixels
            -- warp pipe check!
            if m.controller.buttonDown & D_JPAD ~= 0 and (tile == TILE_PIPE_ENTER_LEFT or tile == TILE_PIPE_WARP_LEFT) and not rightCollision then
                local tile2, px2 = get_tile_at_mario_collider(m, collider_offsets[COLLIDER_RIGHT_FOOT])
                if tile2 == TILE_PIPE_ENTER_RIGHT or tile2 == TILE_PIPE_WARP_RIGHT then
                    -- land on tile, but don't set land flag
                    m.pos.y = truncate(m.pos.y+PX * (PX - px), PX) + m.pos.y % 1
                    m.vel.y = 0
                    set_mario_action(m, ACT_SMB_IN_PIPE, 0)
                    if tile == TILE_PIPE_WARP_LEFT then
                        m.actionArg = 2
                    end
                    return
                end
            end

            -- land on tile/platform
            if smb.platform then
                local o = smb.platform
                m.pos.y = o.oPosY + o.hitboxHeight - o.hitboxDownOffset -- stick to platform
            else
                m.pos.y = truncate(m.pos.y+PX * (PX - px), PX) + m.pos.y % 1
            end
            m.vel.y = 0
            landed = true
        else
            -- djui_chat_message_create(tostring(px)..string.format(", 0x%02x", -floor_or_ceil(m.vel.y)))
            push_out_of_block(m, rightCollision)
            hitWall = true
        end
    end

    -- side check
    usedCollider = COLLIDER_TOP_LEFT
    tile, px = 0, 0
    topPoint = (NES_HEIGHT - 32) * PX -- this doesn't change with size
    rightCollision = false
    if (not smb.isBig) or smb.crouch or m.pos.y + 2 * BLOCK <= topPoint then
        tile, px = get_tile_at_mario_collider(m, collider_offsets[usedCollider])
    end
    if tile == 0 and collider_offsets[usedCollider].y ~= collider_offsets[COLLIDER_BOTTOM_LEFT].y then
        usedCollider = COLLIDER_BOTTOM_LEFT
        tile, px = get_tile_at_mario_collider(m, collider_offsets[usedCollider])
    end
    if tile == 0 then
        rightCollision = true
        usedCollider = COLLIDER_TOP_RIGHT
        if (not smb.isBig) or smb.crouch or m.pos.y + 2 * BLOCK <= topPoint then
            tile, px = get_tile_at_mario_collider(m, collider_offsets[usedCollider])
        end
    end
    if tile == 0 and collider_offsets[usedCollider].y ~= collider_offsets[COLLIDER_BOTTOM_RIGHT].y then
        rightCollision = true
        usedCollider = COLLIDER_BOTTOM_RIGHT
        tile, px = get_tile_at_mario_collider(m, collider_offsets[usedCollider])
    end
    if coin_check(m, usedCollider, tile) then
        return landed
    elseif tile ~= 0 and is_tile_solid(tile) then
        -- side warp pipe check!
        if rightCollision and m.faceAngle.y > 0 and (tile == TILE_PIPE_SIDEWAYS_TL or tile == TILE_PIPE_SIDEWAYS_BL) and usedCollider == COLLIDER_TOP_RIGHT and (m.action ~= ACT_SMB_SWIMMING and m.action & ACT_FLAG_AIR == 0) then
            set_mario_action(m, ACT_SMB_ENTER_SIDE_PIPE, 1)
            return
        end

        push_out_of_block(m, rightCollision)
        hitWall = true
    end

    return landed, hitWall
end

function push_out_of_block(m, rightCollision)
    -- ignore collision in opposite direction
    if floor_or_ceil(m.vel.x) ~= 0 and (floor_or_ceil(m.vel.x) > 0) ~= rightCollision then return false end
    local sign = 1
    if rightCollision then sign = -1 end
    m.pos.x = m.pos.x + sign * PX
    m.vel.x = m.vel.x % 1
    if m.playerIndex == 0 then
        local smb = gSMBStates[0]
        prevMPosForScroll = (m.pos.x * SSPX - 8) - smb.cameraX
    end
end

-- mario's interaction with objects
function smb_mario_object_interactions(m)
    if is_player_active_and_synced(m) == 0 or m.action & ACT_FLAG_INTANGIBLE ~= 0 then return end

    local smb = gSMBStates[m.playerIndex]
    local gIndex = network_global_index_from_local(m.playerIndex)
    m.marioObj.hitboxRadius = 5 * PX -- 5 on both sides of mario
    m.marioObj.hitboxHeight = 12 * PX
    if smb.isBig then
        if smb.crouch then
            m.marioObj.hitboxRadius = 6 * PX 
        else
            m.marioObj.hitboxHeight = 25 * PX
        end
    end

    local marioHurt = false
    local didInteraction = false
    for i,objList in ipairs(VALID_OBJ_LISTS) do
        local o = obj_get_first(objList)
        while o do
            local o_id = get_id_from_behavior(o.behavior)
            if OBJECT_SPRITE_INFO[o_id] and o.oInteractStatus & INT_STATUS_INTERACTED == 0 and o.oIntangibleTimer == 0 and (o.oIntangibleToMarioTimer == 0 or o.globalPlayerIndex ~= gIndex) and (gServerSettings.playerInteractions == PLAYER_INTERACTIONS_PVP or get_id_from_behavior(o.behavior) ~= id_bhvSmbFireball) and mario_and_object_hitbox_overlap(m, o) then
                if (o.oInteractType == INTERACT_BOUNCE_TOP or o.oInteractType == INTERACT_BOUNCE_TOP2 or o.oInteractType == INTERACT_DAMAGE or (o.oInteractType == INTERACT_KOOPA_SHELL and smb.starTimer ~= 0)) then
                    if m.invincTimer == 0 then
                        if (m.playerIndex == 0 or o.allowRemoteInteractions ~= 0) then
                            o.oInteractStatus = o.oInteractStatus | INT_STATUS_INTERACTED
                        end
                        didInteraction = true
                    end

                    if smb.starTimer ~= 0 then
                        if o.oSkipObjectInteraction == 0 and (m.playerIndex == 0 or o.allowRemoteInteractions ~= 0) then
                            o.oInteractStatus = o.oInteractStatus | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED | ATTACK_FAST_ATTACK
                            o.oHealth = 1

                            local points = 2
                            if o.oFireballOverridePoints == 1 then
                                points = o.oDamageOrCoinValue
                            end
                            award_points(m, points, true, o.oPosX, o.oPosY)

                            if m.playerIndex == 0 and o.oSyncID ~= 0 then
                                network_send_object(o, false)
                            end
                        end
                        didInteraction = true
                        --break
                    elseif ((o.oInteractType == INTERACT_BOUNCE_TOP or o.oInteractType == INTERACT_BOUNCE_TOP2) and m.action & ACT_FLAG_SWIMMING == 0 and m.action & ACT_FLAG_AIR ~= 0)
                    and (floor_or_ceil(m.vel.y) < 0 or m.pos.y - 4 * PX >= o.oPosY) then -- always hit if moving down, or hit if 4 pixels above enemy
                        if (m.playerIndex == 0 or o.allowRemoteInteractions ~= 0) then
                            o.oInteractStatus = o.oInteractStatus | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED | ATTACK_FROM_ABOVE
                            if m.playerIndex == 0 and o.oSyncID ~= 0 then
                                network_send_object(o, false)
                            end
                        end
                        didInteraction = true
                        if o.oInteractType == INTERACT_BOUNCE_TOP then
                            m.vel.y = STOMP_VEL + m.vel.y % PX
                            smb.combo = smb.combo + 1
                            award_points(m, smb.combo, true, o.oPosX, o.oPosY)
                        else
                            m.vel.y = STOMP_VEL_OTHER + m.vel.y % PX
                            award_points(m, o.oDamageOrCoinValue, true, o.oPosX, o.oPosY)
                        end
                    elseif not marioHurt then
                        marioHurt = true
                        
                        if m.invincTimer == 0 and (m.playerIndex == 0 or o.allowRemoteInteractions ~= 0) then
                            o.oInteractStatus = o.oInteractStatus | INT_STATUS_INTERACTED
                            if o.oTurnTowardsOnHit == 1 then
                                local tSign = sign(m.pos.x - o.oPosX)
                                o.oMoveAngleYaw = tSign * 0x4000
                                o.oVelX = math.abs(o.oVelX) * tSign
                            end

                            if m.playerIndex == 0 and o.oSyncID ~= 0 then
                                network_send_object(o, false)
                            end
                        end
                    end
                elseif o.oInteractType == INTERACT_POWERUP then -- used for powerups and the axe
                    if (m.playerIndex == 0 or o.allowRemoteInteractions ~= 0) then
                        o.oInteractStatus = o.oInteractStatus | INT_STATUS_INTERACTED
                        if o.oDamageOrCoinValue == 4 then
                            -- do nothing
                        elseif o.oDamageOrCoinValue == 3 then -- 1up
                            award_points(m, 11, true, o.oPosX, o.oPosY)
                        else
                            award_points(m, 6, true, o.oPosX, o.oPosY)
                            if o.oDamageOrCoinValue == 0 or o.oDamageOrCoinValue == 1 then
                                -- mushroom and fire flower
                                local newPower = math.min(smb.powerup + 1, 2)
                                if gGlobalSyncTable.modernPowerStyle then
                                    newPower = o.oDamageOrCoinValue + 1
                                end
                                if smb.powerup < newPower then
                                    smb.powerup = newPower
                                    set_mario_action(m, ACT_SMB_POWER_UP, 0)
                                else
                                    play_smb_sound_stream(m.pos, "powerup")
                                end
                            elseif o.oDamageOrCoinValue == 2 then -- star
                                play_smb_sound_stream(m.pos, "powerup")
                                smb.starTimer = 0x23 * 21
                            end
                        end
                        if m.playerIndex == 0 and o.oSyncID ~= 0 then
                            network_send_object(o, false)
                        end
                    end

                    didInteraction = true
                elseif o.oInteractType == INTERACT_KOOPA_SHELL then -- used for shells, of course
                    if (m.playerIndex == 0 or o.allowRemoteInteractions ~= 0) then
                        o.oInteractStatus = o.oInteractStatus | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED | ATTACK_KICK_OR_TRIP
                        local tSign = sign(o.oPosX - m.pos.x)
                        o.oMoveAngleYaw = tSign * 0x4000
                        o.globalPlayerIndex = gIndex
                        o.oVelX = math.abs(o.oVelX) * tSign
                        o.oPosX = m.pos.x + (m.marioObj.hitboxRadius + o.hitboxRadius) * tSign
                        award_points(m, smb.combo + 3, true, o.oPosX, o.oPosY)
                        if m.playerIndex == 0 and o.oSyncID ~= 0 then
                            network_send_object(o, false)
                        end
                    end

                    didInteraction = true
                elseif o.oInteractType == INTERACT_SPRINGBOARD then -- used for springboards and NOTHING else
                    if m.action ~= ACT_SMB_SPRINGBOARD and m.vel.y < 0 and (m.playerIndex == 0 or o.allowRemoteInteractions ~= 0) then
                        o.oInteractStatus = o.oInteractStatus | INT_STATUS_INTERACTED
                        set_mario_action(m, ACT_SMB_SPRINGBOARD, 0)
                        if m.playerIndex == 0 and o.oSyncID ~= 0 then
                            network_send_object(o, false)
                        end
                    end

                    didInteraction = true
                elseif o.oInteractType == INTERACT_POLE then -- used for... vines. Not the flagpole.
                    if (m.playerIndex == 0 or o.allowRemoteInteractions ~= 0) then
                        if m.action & ACT_FLAG_ON_POLE ~= 0 then
                            m.actionState = 0 -- keep mario on the vine
                            m.usedObj = o
                        else
                            set_mario_action(m, ACT_SMB_CLIMBING, 0)
                            m.usedObj = o
                        end
                    end

                    didInteraction = true
                end
                --break
            end
            o = obj_get_next(o)
        end
    end

    -- player interaction
    if (not didInteraction) and gServerSettings.playerInteractions ~= PLAYER_INTERACTIONS_NONE and m.bounceSquishTimer == 0 and (m.invincTimer == 0 or gServerSettings.playerInteractions ~= PLAYER_INTERACTIONS_PVP) then
        for i=0,MAX_PLAYERS-1 do
            local m2 = gMarioStates[i]
            if i ~= m.playerIndex and is_player_active_and_synced(m2) ~= 0 and m2.action & ACT_FLAG_INTANGIBLE == 0 and m2.bounceSquishTimer == 0 and (m2.invincTimer == 0 or gServerSettings.playerInteractions ~= PLAYER_INTERACTIONS_PVP) then
                local cm2 = (m2.playerIndex == 0 and lag_compensation_get_local_state(gNetworkPlayers[m.playerIndex])) or m2
                local o = m2.marioObj

                -- set other hitbox
                local smb2 = gSMBStates[m2.playerIndex] -- wrong game lol
                o.hitboxRadius = 5 * PX -- 5 on both sides of mario
                o.hitboxHeight = 12 * PX
                if smb2.isBig then
                    if smb2.crouch then
                        o.hitboxRadius = 6 * PX
                    else
                        o.hitboxHeight = 25 * PX
                    end
                end

                -- pvp uses the lag compensation system, but the side interaction does not
                local pvpInteract = mario_and_mario_hitbox_overlap(m, cm2)
                local sideInteract = pvpInteract
                if m2 ~= cm2 then
                    sideInteract = mario_and_mario_hitbox_overlap(m, m2)
                end

                if pvpInteract then
                    didInteraction = true

                    -- star interaction
                    local star1 = (smb.starTimer ~= 0)
                    local star2 = (smb2.starTimer ~= 0)
                    if star1 ~= star2 and gServerSettings.playerInteractions == PLAYER_INTERACTIONS_PVP then
                        if star2 then
                            marioHurt = true
                        else
                            hurt_mario(m2)
                        end
                        break
                    end

                    -- the one moving down gets the hit
                    local p1Hit = floor_or_ceil(m.vel.y) < 0
                    local p2Hit = floor_or_ceil(cm2.vel.y) < 0
                    if p1Hit and p2Hit then -- if both are moving down, check position
                        -- note: these will never both be true
                        p1Hit = m.pos.y - 4 * PX >= cm2.pos.y
                        p2Hit = (not p1Hit) and (cm2.pos.y - 4 * PX >= m.pos.y)
                    end

                    -- side interaction
                    if not (p1Hit or p2Hit) then
                        if not sideInteract then break end
                        local midPointPos = (m.pos.x + m2.pos.x) // 2
                        if m.pos.x >= m2.pos.x then
                            m.pos.x = midPointPos + m.marioObj.hitboxRadius
                            m2.pos.x = midPointPos - m2.marioObj.hitboxRadius
                        else
                            m.pos.x = midPointPos - m.marioObj.hitboxRadius
                            m2.pos.x = midPointPos + m2.marioObj.hitboxRadius
                        end
                        break
                    end

                    local attacker = (p1Hit and m) or cm2
                    local victim = (p2Hit and m) or cm2
                    attacker.vel.y = STOMP_VEL + attacker.vel.y % PX
                    play_smb_sound_stream(victim.pos, "stomp")
                    -- attacker inherits jump height, victim gets bonked
                    if victim.vel.y >= 0 then
                        if attacker.action & ACT_FLAG_SWIMMING == 0 then
                            attacker.vel.y = attacker.vel.y + victim.vel.y
                        else
                            attacker.vel.y = 0
                            set_mario_action(attacker, ACT_SMB_SWIMMING, 1)
                        end
                        victim.vel.y = -(victim.vel.y % PX) - 0x10
                    elseif attacker.action & ACT_FLAG_SWIMMING ~= 0 then
                        attacker.vel.y = 0
                        set_mario_action(attacker, ACT_SMB_SWIMMING, 1)
                    end

                    -- hurt in pvp, or set bounce timer to prevent double interaction
                    -- note that while you can stomp players underwater, it won't deal damage
                    if attacker.action & ACT_FLAG_SWIMMING == 0 and gServerSettings.playerInteractions == PLAYER_INTERACTIONS_PVP then
                        if p2Hit then
                            marioHurt = true
                        else
                            hurt_mario(m2)
                        end
                    end
                    victim.bounceSquishTimer = 5

                    break
                elseif sideInteract then -- side interaction, if pvp is inactive
                    local midPointPos = (m.pos.x + m2.pos.x) // 2
                    if m.pos.x >= m2.pos.x then
                        m.pos.x = midPointPos + m.marioObj.hitboxRadius
                        m2.pos.x = midPointPos - m2.marioObj.hitboxRadius
                    else
                        m.pos.x = midPointPos - m.marioObj.hitboxRadius
                        m2.pos.x = midPointPos + m2.marioObj.hitboxRadius
                    end
                    break
                end
            end
        end
    end

    if marioHurt then
        if m.invincTimer == 0 then
            hurt_mario(m)
        elseif m.invincTimer <= 2 then
            m.invincTimer = m.invincTimer + 2
        end
    end
end

-- interactions when hitting a block
function hit_from_below_interactions(m, blockX, blockY)
    if is_player_active_and_synced(m) == 0 then return end

    local didInteraction = false
    for i,objList in ipairs(VALID_OBJ_LISTS) do
        local o = obj_get_first(objList)
        while o do
            local o_id = get_id_from_behavior(o.behavior)
            if OBJECT_SPRITE_INFO[o_id] and o.oInteractStatus & INT_STATUS_INTERACTED == 0 and o.oIntangibleTimer == 0 and o.oSkipObjectInteraction ~= 1 then
                local x, y = blockX * BLOCK, (blockY+2) * BLOCK
                if object_rect_hitbox_overlap(o, x, y, BLOCK, BLOCK) then
                    if m.playerIndex == 0 then
                        o.oInteractStatus = o.oInteractStatus | INT_STATUS_INTERACTED | INT_STATUS_WAS_ATTACKED | ATTACK_FROM_BELOW
                        if o.oTurnTowardsOnHit == 1 then
                            local tSign = sign((x + BLOCK / 2) - o.oPosX)
                            o.oMoveAngleYaw = tSign * 0x4000
                            o.oVelX = math.abs(o.oVelX) * tSign
                        end
                    end
                    didInteraction = true
                    break
                end
            end
            o = obj_get_next(o)
        end
    end

    -- player interaction
    if (not didInteraction) then
        for i=0,MAX_PLAYERS-1 do
            local m2 = gMarioStates[i]
            if i ~= m.playerIndex and is_player_active_and_synced(m2) ~= 0 and m2.action & ACT_FLAG_INTANGIBLE == 0 and m2.bounceSquishTimer == 0 then
                local x, y = blockX * BLOCK, (blockY+2) * BLOCK
                if mario_rect_hitbox_overlap(m2, x, y, BLOCK, BLOCK) then
                    didInteraction = true
                    -- make victim jump a bit
                    if m2.vel.y >= 0 then
                        m2.vel.y = STOMP_VEL + m.vel.y % PX
                    end

                    -- hurt in pvp and set bounce timer to prevent double interaction
                    m2.bounceSquishTimer = 5
                    if m2.invincTimer == 0 and gServerSettings.playerInteractions == PLAYER_INTERACTIONS_PVP then
                        hurt_mario(m2)
                    end

                    break
                end
            end
        end
    end
end

-- colliders are relative to mario's position. Mario's position is relative to the bottom-middle of the sprite, unlike in super mario bros
COLLIDER_HEAD = 1
COLLIDER_LEFT_FOOT = 2
COLLIDER_RIGHT_FOOT = 3
COLLIDER_TOP_LEFT = 4
COLLIDER_TOP_RIGHT = 5
COLLIDER_BOTTOM_LEFT = 6
COLLIDER_BOTTOM_RIGHT = 7
function get_collider_offsets(m)
    local smb = gSMBStates[m.playerIndex]
    local isBig = (smb.isBig and not smb.crouch)
    local collider_offsets = {}
    collider_offsets[COLLIDER_HEAD] = {x = 0, y = 14, type = COLLIDER_HEAD} -- head
    collider_offsets[COLLIDER_LEFT_FOOT] = {x = -5, y = 0, type = COLLIDER_LEFT_FOOT} -- left foot
    collider_offsets[COLLIDER_RIGHT_FOOT] = {x = 4, y = 0, type = COLLIDER_RIGHT_FOOT} -- right foot
    collider_offsets[COLLIDER_TOP_LEFT] = {x = -6, y = 8, type = COLLIDER_TOP_LEFT} -- top left side
    collider_offsets[COLLIDER_TOP_RIGHT] = {x = 5, y = 8, type = COLLIDER_TOP_RIGHT} -- top right side
    collider_offsets[COLLIDER_BOTTOM_LEFT] = {x = -6, y = 8, type = COLLIDER_BOTTOM_LEFT} -- bottom left side
    collider_offsets[COLLIDER_BOTTOM_RIGHT] = {x = 5, y = 8, type = COLLIDER_BOTTOM_RIGHT} -- bottom right side

    -- adjust for big mario
    if isBig then
        collider_offsets[COLLIDER_HEAD].y = 28
        collider_offsets[COLLIDER_TOP_LEFT].y = 24
        collider_offsets[COLLIDER_TOP_RIGHT].y = 24
    end

    return collider_offsets
end

-- handle interactions with coins
function coin_check(m, collider, tile)
    if tile == TILE_COIN then
        if m.playerIndex == 0 then
            local collider_offsets = get_collider_offsets(m)
            local blockX, blockY = get_block_coords_for_mario_collider(m, collider_offsets[collider])
            network_send_to_all_in_act(true, {
                id = PACKET_BLOCK_INTERACT,
                x = blockX,
                y = blockY,
                tile = tile,
                area = gNetworkPlayers[0].currAreaIndex,
            })
        end
        return true
    end
end

-- hurt routine
function hurt_mario(m, death)
    if m.playerIndex ~= 0 then return end -- don't run for remote players
    local smb = gSMBStates[m.playerIndex]
    if smb.powerup == 0 or death then
        m.numLives = m.numLives - 1
        if mario_can_bubble_smb(m) then
            smb.powerup = 0
            smb.isBig = false
            set_mario_action(m, ACT_SMB_BUBBLED, 0)
        else
            set_mario_action(m, ACT_SMB_DEAD, (death and 1) or 0)
        end
    elseif gGlobalSyncTable.modernPowerStyle then
        smb.powerup = smb.powerup - 1
    else
        smb.powerup = 0
    end
    if m.action ~= ACT_SMB_DEAD and m.action ~= ACT_SMB_BUBBLED then
        set_mario_action(m, ACT_SMB_DAMAGE, 0)
    end
end

-- point routine
local POINT_CONVERT = {100, 200, 400, 500, 800, 1000, 2000, 4000, 5000, 8000} -- after is 1up!
function award_points(m, points, visualEffect, effectPosX, effectPosY, flagPole)
    local smb = gSMBStates[m.playerIndex]
    if points > #POINT_CONVERT then
        points = #POINT_CONVERT + 1
        m.numLives = m.numLives + 1
        if m.playerIndex == 0 then
            play_smb_sound_stream(m.pos, "1up")
        end
    else
        smb.points = smb.points + POINT_CONVERT[points]
    end

    -- only show visual effect for local players
    if m.playerIndex == 0 and visualEffect then
        spawn_object_no_rotate(id_bhvSmbPointNumber, E_MODEL_NONE, effectPosX, effectPosY + BLOCK, 0, false, function(o)
            o.oAnimState = points - 1
            if flagPole then
                o.oAction = 1
            end
        end)
    end
end

--- physics constants ---
MIN_WALK_VEL = 0x130 * SSSPX
MAX_WALK_VEL = 0x1900 * SSSPX
MAX_WALK_VEL_WATER = 0x1100 * SSSPX
MAX_WALK_VEL_ENTRY = 0xD00 * SSSPX
MAX_RUN_VEL = 0x2900 * SSSPX
SKID_TURN_VEL = 0x900 * SSSPX
WALK_ACCEL = 0x98 * SSSPX
RUN_ACCEL = 0xE4 * SSSPX
SKID_ACCEL = 0xD0 * SSSPX -- acceleration when not moving or skidding

JUMP_ACCEL = 0x98 * SSSPX -- used when holding forward while jumping or holding back from a slow jump
JUMP_ACCEL_FAST = 0xE4 * SSSPX -- used when holding forward or back while jumping with some speed
JUMP_ACCEL_FAST_BACK = 0xD0 * SSSPX -- used when holding back from a fast jump

JUMP_VEL = 0x4000 * SSSPX
JUMP_VEL_FAST = 0x5000 * SSSPX
STOMP_VEL = 0x4000 * SSSPX -- goombas, koopas, buzzy beetles
STOMP_VEL_OTHER = 0x3000 * SSSPX -- lakitus, hammer bros, cheep-cheeps, bullet bills

ENTRY_GRAVITY = 0x280 * SSSPX -- initial gravity when starting a level
HOLD_A_GRAVITY_SLOW = 0x200 * SSSPX -- from slow jump
FALL_GRAVITY_SLOW = 0x700 * SSSPX
HOLD_A_GRAVITY = 0x01E0 * SSSPX
FALL_GRAVITY = 0x600 * SSSPX
HOLD_A_GRAVITY_FAST = 0x280 * SSSPX -- from fast jump
FALL_GRAVITY_FAST = 0x900 * SSSPX
TERMINAL_VEL = 0x4800 * SSSPX -- terminal velocity

SWIM_VEL = 0x1800 * SSSPX -- initial vel when swimming upwards
SWIM_VEL_WHIRLPOOL = 0x1000 * SSSPX -- initial vel when swimming upwards in a whirlpool
TERMINAL_VEL_WHIRLPOOL = 0x1800 * SSSPX -- terminal velocity in a whirlpool

HOLD_A_GRAVITY_WATER = 0xD0 * SSSPX
FALL_GRAVITY_WATER = 0xA0 * SSSPX
HOLD_A_GRAVITY_WATER_WHIRLPOOL = 0x40 * SSSPX -- gravity in whirlpool
FALL_GRAVITY_WATER_WHIRLPOOL = 0x90 * SSSPX
GRAVITY_WATER_SURFACE = 0x180 * SSSPX -- gravity when near the water surface (applies for both falling and holding A)

-- standing still
function act_smb_idle(m)
    local smb = gSMBStates[m.playerIndex]
    if m.controller.buttonPressed & A_BUTTON ~= 0 then
        return set_mario_action(m, ACT_SMB_JUMP, 0)
    elseif floor_or_ceil(m.vel.y) ~= 0 then
        return set_mario_action(m, ACT_SMB_FALL, 0)
    elseif m.controller.buttonDown & D_JPAD ~= 0 then
        smb.crouch = true
    elseif m.vel.x ~= 0 or (m.controller.buttonDown & (L_JPAD | R_JPAD) ~= 0) then
        smb.crouch = false
        return set_mario_action(m, ACT_SMB_WALKING, 0)
    else
        smb.crouch = false
    end

    m.vel.y = smb_vertical_acceleration(m, smb.gravity)
    smb_mario_step(m)
    
    return 0
end
ACT_SMB_IDLE = allocate_and_hook_mario_action(act_smb_idle, ACT_GROUP_STATIONARY | ACT_FLAG_STATIONARY)

-- walking, running, and skidding
function act_smb_walking(m)
    local smb = gSMBStates[m.playerIndex]
    if m.controller.buttonPressed & A_BUTTON ~= 0 then
        return set_mario_action(m, ACT_SMB_JUMP, 0)
    elseif floor_or_ceil(m.vel.y) ~= 0 then
        return set_mario_action(m, ACT_SMB_FALL, 0)
    end

    if m.controller.buttonDown & D_JPAD ~= 0 then
        smb.crouch = true
    else
        smb.crouch = false
    end

    m.vel.y = smb_vertical_acceleration(m, smb.gravity)

    local approachSpeed = MAX_WALK_VEL
    local accel = WALK_ACCEL
    if m.controller.buttonDown & B_BUTTON ~= 0 then
        m.actionState = 10
        approachSpeed = MAX_RUN_VEL
        accel = RUN_ACCEL
    elseif m.actionState ~= 0 then
        m.actionState = m.actionState - 1
        approachSpeed = MAX_RUN_VEL
        accel = RUN_ACCEL
    end

    if m.controller.buttonDown & D_JPAD ~= 0 then
        approachSpeed = 0
        accel = SKID_ACCEL
    elseif m.controller.buttonDown & R_JPAD ~= 0 then
        m.faceAngle.y = 0x4000
        if m.vel.x < -SKID_TURN_VEL then
            accel = SKID_ACCEL
        end
    elseif m.controller.buttonDown & L_JPAD ~= 0 then
        approachSpeed = -approachSpeed
        m.faceAngle.y = -0x4000
        if m.vel.x > SKID_TURN_VEL then
            accel = SKID_ACCEL
        end
    else
        approachSpeed = 0
        accel = SKID_ACCEL
    end
    
    m.actionTimer = m.actionTimer + 1
    local changeTime = 8
    if math.abs(m.vel.x) >= MAX_RUN_VEL - 1 then
        changeTime = 2
    elseif math.abs(m.vel.x) >= MAX_WALK_VEL - 1 then
        changeTime = 4
    end
    if m.vel.x ~= 0 and m.actionTimer >= changeTime then
        smb.walkCycleFrame = (smb.walkCycleFrame + 1) % 3
        m.actionTimer = 0
    end
    m.vel.x = smb_horizontal_acceleration(m, accel, approachSpeed)
    smb_mario_step(m)
    
    if floor_or_ceil(m.vel.x) == 0 and (m.controller.buttonDown & (L_JPAD | R_JPAD) == 0) then
        set_mario_action(m, ACT_SMB_IDLE, 0)
    end
    return 0
end
ACT_SMB_WALKING = allocate_and_hook_mario_action(act_smb_walking, ACT_GROUP_MOVING | ACT_FLAG_MOVING)

-- jumping
function act_smb_jump(m)
    local smb = gSMBStates[m.playerIndex]
    if not smb.isBig then
        play_mario_action_sound_stream(m, "jumpSmall")
    else
        play_mario_action_sound_stream(m, "jumpBig")
    end
    
    local accel = 0
    local holdDir = 0
    if m.controller.buttonDown & R_JPAD ~= 0 then
        holdDir = 1
    elseif m.controller.buttonDown & L_JPAD ~= 0 then
        holdDir = -1
    end

    local currDir = sign(m.faceAngle.y)
    if holdDir == currDir then -- holding forward
        if math.abs(m.vel.x) >= MAX_WALK_VEL then
            accel = JUMP_ACCEL_FAST
        else
            accel = JUMP_ACCEL
        end
    elseif holdDir ~= 0 then -- holding backward
        if math.abs(m.vel.x) >= MAX_WALK_VEL then
            accel = JUMP_ACCEL_FAST
        elseif math.abs(smb.initialJumpXVel) >= 0x1D then
            accel = JUMP_ACCEL_FAST_BACK
        else
            accel = JUMP_ACCEL
        end
    end

    local approachSpeed = 0 -- maintain speed (0 is ignored)
    if holdDir == 0 then
        -- do nothing
    elseif math.abs(smb.initialJumpXVel) <= holdDir * MAX_WALK_VEL then
        approachSpeed = holdDir * MAX_WALK_VEL
    else
        approachSpeed = holdDir * MAX_RUN_VEL
    end
    
    --djui_chat_message_create(tostring(m.vel.x)..":"..tostring(accel))
    
    local gravity = 0
    m.actionTimer = m.actionTimer + 1
    if m.vel.y < 0 then
        m.actionState = 1
    end
    if m.actionState == 0 and (m.controller.buttonDown & A_BUTTON ~= 0) then
        -- hold A gravity based on initial jump speed
        gravity = smb.holdAGravity
    else
        gravity = smb.gravity
    end
    m.vel.y = smb_vertical_acceleration(m, gravity)

    if approachSpeed ~= 0 then
        m.vel.x = smb_horizontal_acceleration(m, accel, approachSpeed)
    end
    local landed = smb_mario_step(m)
    
    if landed then
        -- this quirk exists with the original game too
        if smb.platform == nil then
            smb.combo = 0
        end
        set_mario_action(m, ACT_SMB_WALKING, 0)
    end
    return 0
end
ACT_SMB_JUMP = allocate_and_hook_mario_action(act_smb_jump, ACT_GROUP_AIRBORNE | ACT_FLAG_MOVING | ACT_FLAG_AIR | ACT_FLAG_CONTROL_JUMP_HEIGHT)

-- falling
function act_smb_fall(m)
    local smb = gSMBStates[m.playerIndex]
    local accel = MAX_WALK_VEL
    local holdDir = 0
    if m.controller.buttonDown & R_JPAD ~= 0 then
        holdDir = 1
    elseif m.controller.buttonDown & L_JPAD ~= 0 then
        holdDir = -1
    end

    -- break bubble sfx
    if m.actionArg == 2 then
        play_mario_action_sound_stream(m, "break")
    end

    local currDir = sign(m.faceAngle.y)
    if holdDir == currDir then -- holding forward
        if math.abs(m.vel.x) >= MAX_WALK_VEL then
            accel = JUMP_ACCEL_FAST
        else
            accel = JUMP_ACCEL
        end
    elseif holdDir ~= 0 then -- holding backward
        if math.abs(m.vel.x) >= MAX_WALK_VEL then
            accel = JUMP_ACCEL_FAST
        elseif math.abs(smb.initialJumpXVel) >= 0x1D then
            accel = JUMP_ACCEL_FAST_BACK
        else
            accel = JUMP_ACCEL
        end
    end

    local approachSpeed = 0 -- maintain speed (0 is ignored)
    if holdDir == 0 then
        -- do nothing
    elseif math.abs(smb.initialJumpXVel) < holdDir * MAX_WALK_VEL then
        approachSpeed = holdDir * MAX_WALK_VEL
    else
        approachSpeed = holdDir * MAX_RUN_VEL
    end
    
    --djui_chat_message_create(tostring(m.vel.x)..":"..tostring(accel))

    m.actionTimer = m.actionTimer + 1
    m.vel.y = smb_vertical_acceleration(m, smb.gravity)

    if approachSpeed ~= 0 then
        m.vel.x = smb_horizontal_acceleration(m, accel, approachSpeed)
    end
    local landed = smb_mario_step(m)
    
    if landed then
        -- this quirk exists with the original game too
        if smb.platform == nil then
            smb.combo = 0
        end
        set_mario_action(m, ACT_SMB_WALKING, 0)
    end
    return 0
end
ACT_SMB_FALL = allocate_and_hook_mario_action(act_smb_fall, ACT_GROUP_AIRBORNE | ACT_FLAG_MOVING | ACT_FLAG_AIR)

-- standing still underwater
function act_smb_water_idle(m)
    local smb = gSMBStates[m.playerIndex]
    if m.controller.buttonPressed & A_BUTTON ~= 0 then
        return set_mario_action(m, ACT_SMB_SWIMMING, 1)
    elseif floor_or_ceil(m.vel.y) ~= 0 then
        return set_mario_action(m, ACT_SMB_SWIMMING, 0)
    elseif m.controller.buttonDown & D_JPAD ~= 0 then
        smb.crouch = true
    elseif m.vel.x ~= 0 or (m.controller.buttonDown & (L_JPAD | R_JPAD) ~= 0) then
        smb.crouch = false
        return set_mario_action(m, ACT_SMB_WATER_WALKING, 0)
    else
        smb.crouch = false
    end

    m.vel.y = smb_vertical_acceleration(m, smb.gravity)
    smb_mario_step(m)
    
    return 0
end
ACT_SMB_WATER_IDLE = allocate_and_hook_mario_action(act_smb_water_idle, ACT_GROUP_STATIONARY | ACT_FLAG_STATIONARY | ACT_FLAG_SWIMMING | ACT_FLAG_SWIMMING_OR_FLYING | ACT_FLAG_WATER_OR_TEXT)

-- walking underwater
function act_smb_water_walking(m)
    local smb = gSMBStates[m.playerIndex]
    if m.controller.buttonPressed & A_BUTTON ~= 0 then
        return set_mario_action(m, ACT_SMB_SWIMMING, 1)
    elseif floor_or_ceil(m.vel.y) ~= 0 then
        return set_mario_action(m, ACT_SMB_SWIMMING, 0)
    end

    if m.controller.buttonDown & D_JPAD ~= 0 then
        smb.crouch = true
    else
        smb.crouch = false
    end

    m.vel.y = smb_vertical_acceleration(m, smb.gravity)

    local approachSpeed = MAX_WALK_VEL_WATER
    local accel = WALK_ACCEL

    if m.controller.buttonDown & D_JPAD ~= 0 then
        approachSpeed = 0
        accel = SKID_ACCEL
    elseif m.controller.buttonDown & R_JPAD ~= 0 then
        m.faceAngle.y = 0x4000
        if m.vel.x < -SKID_TURN_VEL then
            accel = SKID_ACCEL
        end
    elseif m.controller.buttonDown & L_JPAD ~= 0 then
        approachSpeed = -approachSpeed
        m.faceAngle.y = -0x4000
        if m.vel.x > SKID_TURN_VEL then
            accel = SKID_ACCEL
        end
    else
        approachSpeed = 0
        accel = SKID_ACCEL
    end
    
    m.actionTimer = m.actionTimer + 1
    local changeTime = 8
    if m.vel.x ~= 0 and m.actionTimer >= changeTime then
        smb.walkCycleFrame = (smb.walkCycleFrame + 1) % 3
        m.actionTimer = 0
    end
    m.vel.x = smb_horizontal_acceleration(m, accel, approachSpeed)
    smb_mario_step(m)
    
    if floor_or_ceil(m.vel.x) == 0 and (m.controller.buttonDown & (L_JPAD | R_JPAD) == 0) then
        set_mario_action(m, ACT_SMB_WATER_IDLE, 0)
    end
    return 0
end
ACT_SMB_WATER_WALKING = allocate_and_hook_mario_action(act_smb_water_walking, ACT_GROUP_MOVING | ACT_FLAG_MOVING | ACT_FLAG_SWIMMING | ACT_FLAG_SWIMMING_OR_FLYING | ACT_FLAG_WATER_OR_TEXT)

-- swimming action
function act_smb_swimming(m)
    local smb = gSMBStates[m.playerIndex]
    if m.actionArg == 1 then
        play_mario_action_sound_stream(m, "swim")
    elseif m.actionArg == 2 then
        -- break bubble sfx
        play_mario_action_sound_stream(m, "break")
    end

    if m.controller.buttonPressed & A_BUTTON ~= 0 and m.actionTimer ~= 0 then
        return set_mario_action(m, ACT_SMB_SWIMMING, 1)
    end
    
    local accel = 0
    local holdDir = 0
    if m.controller.buttonDown & R_JPAD ~= 0 then
        m.faceAngle.y = 0x4000
        holdDir = 1
    elseif m.controller.buttonDown & L_JPAD ~= 0 then
        m.faceAngle.y = -0x4000
        holdDir = -1
    end

    if math.abs(m.vel.x) >= MAX_WALK_VEL then
        accel = JUMP_ACCEL_FAST
    else
        accel = JUMP_ACCEL
    end

    local approachSpeed = 0 -- maintain speed (0 is ignored)
    if holdDir == 0 then
        -- do nothing
    else
        approachSpeed = holdDir * MAX_WALK_VEL
    end
    
    --djui_chat_message_create(tostring(m.vel.x)..":"..tostring(accel))

    local gravity = 0
    m.actionTimer = m.actionTimer + 1
    if m.vel.y < 0 then
        m.actionState = 1
        smb.walkCycleFrame = 0
    elseif m.actionTimer % 4 == 0 then
        smb.walkCycleFrame = (smb.walkCycleFrame + 1) % 3
    end
    if m.actionState == 0 and (m.controller.buttonDown & A_BUTTON ~= 0) then
        gravity = smb.holdAGravity
    else
        gravity = smb.gravity
    end
    m.vel.y = smb_vertical_acceleration(m, gravity)

    if approachSpeed ~= 0 then
        m.vel.x = smb_horizontal_acceleration(m, accel, approachSpeed)
    end
    local landed = smb_mario_step(m)
    
    if landed then
        -- this quirk exists with the original game too
        if smb.platform == nil then
            smb.combo = 0
        end
        set_mario_action(m, ACT_SMB_WATER_WALKING, 0)
    end
    return 0
end
ACT_SMB_SWIMMING = allocate_and_hook_mario_action(act_smb_swimming, ACT_GROUP_AIRBORNE | ACT_FLAG_MOVING | ACT_FLAG_SWIMMING | ACT_FLAG_SWIMMING_OR_FLYING)

-- when getting a power up
function act_smb_power_up(m)
    local smb = gSMBStates[m.playerIndex]
    if m.playerIndex == 0 then
        custom_handle_time_stop()
    end
    play_mario_action_sound_stream(m, "powerup")

    -- powerup jump
    if m.controller.buttonDown & A_BUTTON == 0 and m.actionTimer == 0 then
        m.actionState = 1
    end

    m.actionTimer = m.actionTimer + 1
    if m.actionTimer == 16 then
        if gGlobalSyncTable.modernPowerStyle then
            smb.isBig = (smb.powerup ~= 0)
        elseif smb.powerup == 1 then
            smb.isBig = not smb.isBig
        end
    elseif m.actionTimer >= 56 then
        if m.playerIndex == 0 and timeStopEnabled then
            disable_time_stop()
            timeStopEnabled = false
        else
            m.invincTimer = 8 * 21
        end
        
        if m.actionState == 1 and m.controller.buttonDown & A_BUTTON ~= 0 then
            set_mario_action(m, ACT_SMB_JUMP, 0)
        else
            set_mario_action(m, ACT_SMB_IDLE, 0)
        end
    end
    return 0
end
ACT_SMB_POWER_UP = allocate_and_hook_mario_action(act_smb_power_up, ACT_GROUP_AUTOMATIC | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE)

-- taking damage
function act_smb_damage(m)
    local smb = gSMBStates[m.playerIndex]
    if m.playerIndex == 0 then
        custom_handle_time_stop()
    end
    m.invincTimer = 8 * 21
    play_mario_action_sound_stream(m, "hurt")

    m.actionTimer = m.actionTimer + 1
    if m.actionTimer == 16 then
        if smb.powerup == 0 then
            smb.isBig = not smb.isBig
        end
    elseif m.actionTimer >= 56 then
        local newAction = m.prevAction
        if m.prevAction == m.action then
            newAction = ACT_SMB_FALL
        end
        if m.playerIndex == 0 then
            disable_time_stop()
            timeStopEnabled = false
        end
        m.action = newAction
        m.actionArg = 0
        m.actionTimer = 0
    end
    return 0
end
ACT_SMB_DAMAGE = allocate_and_hook_mario_action(act_smb_damage, ACT_GROUP_AUTOMATIC | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE)

-- dead (both pit and from enemies)
function act_smb_dead(m)
    if m.playerIndex == 0 and (m.actionArg == 0 or m.actionState == 1) then
        m.actionState = 1
        custom_handle_time_stop()
    end
    m.invincTimer = 0
    if m.playerIndex == 0 then
        set_current_music("death")
        m.flags = m.flags | MARIO_ACTION_SOUND_PLAYED
    else
        play_mario_action_sound_stream(m, "deathShort")
    end

    m.actionTimer = m.actionTimer + 1
    if m.actionArg == 0 then
        if m.actionTimer == 30 then
            m.vel.y = JUMP_VEL
        elseif m.actionTimer > 30 then
            m.vel.y = m.vel.y - ENTRY_GRAVITY
            m.pos.y = m.pos.y + floor_or_ceil(m.vel.y)
            if m.pos.y < -16 * PX then
                m.actionArg = 1
            end
        end
    elseif m.actionTimer >= 240 and m.playerIndex == 0 then
        disable_time_stop()
        timeStopEnabled = false
        entranceType = 0
        lastAct = -1
        if m.numLives <= 0 then
            set_mario_action(m, ACT_SMB_LOADING, 1)
        elseif respawnArea ~= 1 then
            warp_to_level(LEVEL_SMB, respawnArea, gNetworkPlayers[0].currActNum)
        else
            warp_restart_level()
        end
    end
    return 0
end
ACT_SMB_DEAD = allocate_and_hook_mario_action(act_smb_dead, ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE)

-- bubbled
function act_smb_bubbled(m)
    local smb = gSMBStates[m.playerIndex]
    m.invincTimer = 0
    smb.loopScreenLeft, smb.loopScreenRight = 0, 0
    play_mario_action_sound_stream(m, "deathShort")

    local m2 = nearest_non_cutscene_mario_state_to_object(m.marioObj)
    if not m2 then
        if localOptionTable.debugBubble then
            m2 = gMarioStates[1]
        else
            m.faceAngle.y = sign(m.faceAngle.y) * 0x4000
            return set_mario_action(m, ACT_SMB_DEAD, 0)
        end
    end

    if m.actionArg == 0 then
        m.actionTimer = m.actionTimer + 1
        if m.actionTimer > 30 then
            m.actionArg = 1
            m.actionTimer = 0
        end
        return 0
    end

    m.actionTimer = m.actionTimer + 1
    if m.actionState == 0 then
        m.faceAngle.y = m.faceAngle.y + 0x100
        m.vel.x = 10 * sins(m.faceAngle.y)
        m.vel.y = 10 * coss(m.faceAngle.y)
        if m.controller.buttonPressed & A_BUTTON ~= 0 then
            m.faceAngle.y = calculate_yaw({x = m.pos.x, y = 0, z = m.pos.y}, {x = m2.pos.x, y = 0, z = m2.pos.y})
            m.actionState = 1
            m.actionTimer = 0
        end
    else
        m.vel.x = 20 * sins(m.faceAngle.y)
        m.vel.y = 20 * coss(m.faceAngle.y)
        if m.actionTimer > 30 then
            m.actionState = 0
        end
    end
    m.pos.x = m.pos.x + m.vel.x
    m.pos.y = m.pos.y + m.vel.y
    if m.pos.y < 0 then
        m.pos.y = 0
    elseif m.pos.y > NES_HEIGHT * PX - BLOCK then
        m.pos.y = NES_HEIGHT * PX - BLOCK
    end

    if m.playerIndex == 0 and get_dist_between_objects_2d(m.marioObj, m2.marioObj) < BLOCK then
        local smb2 = gSMBStates[m2.playerIndex]
        m.pos.x, m.pos.y = m2.pos.x, m2.pos.y
        smb.loopScreenLeft, smb.loopScreenRight = smb2.loopScreenLeft, smb2.loopScreenRight
        m.invincTimer = 8 * 21
        m.faceAngle.y = sign(m.faceAngle.y) * 0x4000
        m.vel.x = 0
        m.vel.y = JUMP_VEL
        smb.gravity = ENTRY_GRAVITY
        timeUpDeath = false
        return set_mario_action(m, ACT_SMB_FALL, 2)
    end

    if m.playerIndex == 0 then
        smb_move_camera(m, true)
    end

    return 0
end
ACT_SMB_BUBBLED = allocate_and_hook_mario_action(act_smb_bubbled, ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE)

-- going into a vertical pipe, or exiting a vertical pipe
function act_smb_in_pipe(m)
    m.actionTimer = m.actionTimer + 1
    if m.actionArg ~= 1 then -- going down
        play_mario_action_sound_stream(m, "pipe")
        if m.actionTimer < 55 then
            m.pos.y = m.pos.y - 10
        end

        -- do warp
        if m.actionTimer >= 90 and m.playerIndex == 0 then
            local np = gNetworkPlayers[0]
            if m.actionArg == 0 then
                local area = 1
                entranceType = 2
                if pipeEntranceArea ~= 0 then
                    area = pipeEntranceArea
                    entranceType = (doUpwardPipe and entranceType) or 1
                end
                warp_to_level(LEVEL_SMB, area, np.currActNum)
            else -- warp zone
                local world = 0
                -- check mario position on screen
                local screenBlock = (m.pos.x % (BLOCK * 16)) // BLOCK
                if screenBlock < 6 then
                    world = warpZoneData[1]
                elseif screenBlock < 10 then
                    world = warpZoneData[2]
                else
                    world = warpZoneData[3]
                end
                entranceType = 0
                warp_to_level(LEVEL_SMB, 1, world * 4 - 3)
            end
        end
    else -- going up
        m.pos.y = m.pos.y + 10
        if m.actionTimer >= 100 then
            if (network_player_connected_count() > 1) then
                m.invincTimer = 8 * 21
            end
            set_mario_action(m, ACT_SMB_IDLE, 0)
        end
    end
end

ACT_SMB_IN_PIPE = allocate_and_hook_mario_action(act_smb_in_pipe, ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE)

-- going into a sideways pipe (or the intro cutscene in 1-2, 2-2, etc.)
function act_smb_enter_side_pipe(m)
    if m.actionTimer == 0 then
        m.vel.x = MAX_WALK_VEL_ENTRY
    end
    m.actionTimer = m.actionTimer + 1
    if m.actionTimer % 8 == 0 then
        local smb = gSMBStates[m.playerIndex]
        smb.walkCycleFrame = (smb.walkCycleFrame + 1) % 3
    end

    -- actionArg is 0 until we actually enter the pipe, at which point it becomes 1
    if m.actionArg == 0 then
        smb_mario_step(m)
    elseif m.actionArg == 1 then
        play_mario_action_sound_stream(m, "pipe")
        m.pos.x = m.pos.x + floor_or_ceil(m.vel.x)
        if m.actionTimer > 16 then
            m.pos.x = truncate(m.pos.x, BLOCK) + 0.5 * BLOCK
            m.vel.x = 0
        end

        -- do the warp
        if m.playerIndex == 0 and m.actionTimer >= 90 then
            local np = gNetworkPlayers[0]
            local area = respawnArea
            entranceType = 2
            if loaded_tile_data.pipeCutscene then
                entranceType = 1
            end
            if pipeEntranceArea ~= 0 then
                area = pipeEntranceArea
            else
                doBonusExit = true
            end
            warp_to_level(np.currLevelNum, area, np.currActNum)
        end
    end
end

ACT_SMB_ENTER_SIDE_PIPE = allocate_and_hook_mario_action(act_smb_enter_side_pipe, ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE)

-- sliding down the flag at the end of a level
function act_smb_slide_down_flag(m)
    gSMBStates[m.playerIndex].walkCycleFrame = 0
    if m.actionArg == 0 then
        if m.playerIndex == 0 and (m.usedObj == m.marioObj or m.usedObj == nil or m.usedObj.oAction == 2) then
            m.actionArg = 1
            m.pos.x = m.pos.x + BLOCK
            return 0
        end

        play_mario_action_sound_stream(m, "flagpole")
        m.faceAngle.y = 0x4000
        if m.playerIndex == 0 then
            m.pos.x = m.usedObj.oPosX
            set_current_music("")
        end
        m.pos.y = m.pos.y - 2 * PX
        if m.pos.y <= BLOCK * 3 then
            m.pos.y = BLOCK * 3
        end
    elseif m.actionArg == 1 then
        m.faceAngle.y = -0x4000
        m.actionTimer = m.actionTimer + 1
        if m.actionTimer > 16 then
            m.vel.x = 0
            m.vel.y = 0
            set_mario_action(m, ACT_SMB_WALK_CUTSCENE, 0)
        end
    end
    return 0
end

ACT_SMB_SLIDE_DOWN_FLAG = allocate_and_hook_mario_action(act_smb_slide_down_flag, ACT_GROUP_CUTSCENE | ACT_FLAG_ON_POLE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE)

-- bouncing on a spring; pressing A during this action will make you jump higher
function act_smb_springboard(m)
    local smb = gSMBStates[m.playerIndex]
    smb.crouch = false
    if m.controller.buttonDown & R_JPAD ~= 0 then
        m.faceAngle.y = 0x4000
    elseif m.controller.buttonDown & L_JPAD ~= 0 then
        m.faceAngle.y = -0x4000
    end

    if m.controller.buttonPressed & A_BUTTON ~= 0 then
        m.actionState = 1
    end

    m.actionTimer = m.actionTimer + 1
    if m.actionTimer <= 8 then
        m.vel.y = -PX
    else
        m.vel.y = PX
    end
    m.pos.y = m.pos.y + m.vel.y

    if m.actionTimer >= 18 then
        if m.actionState == 0 then
            m.vel.y = 0x7 * PX
        else
            m.vel.y = 0xC * PX
        end
        set_mario_action(m, ACT_SMB_FALL, 0)
    end

    return 0
end
ACT_SMB_SPRINGBOARD = allocate_and_hook_mario_action(act_smb_springboard, ACT_GROUP_AIRBORNE | ACT_FLAG_AIR)

-- climbing a vine
function act_smb_climbing(m)
    local smb = gSMBStates[m.playerIndex]
    smb.crouch = false
    
    m.actionTimer = m.actionTimer + 1
    if m.actionArg == 0 then
        if m.playerIndex == 0 then
            m.pos.x = m.usedObj.oPosX - 8 * PX * sign(m.faceAngle.y)
        end

        if m.controller.buttonDown & R_JPAD ~= 0 then
            if m.faceAngle.y > 0 then
                m.faceAngle.y = -0x4000
                m.actionTimer = 1
            elseif m.actionTimer > 2 and (m.controller.buttonPressed & R_JPAD ~= 0 or m.actionTimer > 25) then
                m.pos.x = m.pos.x + 2 * PX
                return set_mario_action(m, ACT_SMB_FALL, 0)
            end
        elseif m.controller.buttonDown & L_JPAD ~= 0 then
            if m.faceAngle.y <= 0 then
                m.faceAngle.y = 0x4000
                m.actionTimer = 1
            elseif m.actionTimer > 2 and (m.controller.buttonPressed & L_JPAD ~= 0 or m.actionTimer > 25) then
                m.pos.x = m.pos.x - 2 * PX
                return set_mario_action(m, ACT_SMB_FALL, 0)
            end
        end

        if m.controller.buttonDown & U_JPAD ~= 0 then
            m.vel.y = PX
            if m.actionTimer % 8 == 0 then
                smb.walkCycleFrame = (smb.walkCycleFrame + 1) % 2
            end
        elseif m.controller.buttonDown & D_JPAD ~= 0 then
            m.vel.y = -2 * PX
            if m.actionTimer % 4 == 0 then
                smb.walkCycleFrame = (smb.walkCycleFrame + 1) % 2
            end
        else
            smb.walkCycleFrame = 0
            m.vel.y = 0
        end
        m.vel.x = 0
        smb_mario_step(m, true)

        m.actionState = m.actionState + 1
        if m.actionState >= 3 then
            return set_mario_action(m, ACT_SMB_FALL, 0)
        end

        -- switch to cutscene version if we're above a certain point
        if m.pos.y > NES_HEIGHT * PX - 4 * BLOCK then
            m.actionArg = 1
        end
    elseif m.actionArg == 1 then
        if m.playerIndex == 0 then
            m.pos.x = m.usedObj.oPosX - 8 * PX * sign(m.faceAngle.y)
        end

        m.vel.y = PX
        m.vel.x = 0
        if m.actionTimer % 8 == 0 then
            smb.walkCycleFrame = (smb.walkCycleFrame + 1) % 2
        end
        local landed, hitWall = smb_mario_step(m, true)
        if landed then
            smb.walkCycleFrame = 0
        end

        -- do warp
        if m.playerIndex == 0 and m.pos.y > NES_HEIGHT * PX then
            local np = gNetworkPlayers[0]
            local area = 1
            entranceType = 3
            if pipeEntranceArea ~= 0 then
                area = pipeEntranceArea
            end
            warp_to_level(LEVEL_SMB, area, np.currActNum)
        end
    end
    return 0
end

ACT_SMB_CLIMBING = allocate_and_hook_mario_action(act_smb_climbing, ACT_GROUP_OBJECT | ACT_FLAG_ON_POLE)

-- cutscene when entering from a vine
function act_smb_vine_cutscene(m)
    local smb = gSMBStates[m.playerIndex]
    smb.crouch = false

    if m.actionState == 1 then
        if m.faceAngle.y > 0 then
            m.faceAngle.y = -0x4000
            m.actionTimer = 1
        elseif m.actionTimer > 25 then
            m.pos.x = m.pos.x + 2 * PX
            return set_mario_action(m, ACT_SMB_FALL, 0)
        end
    else
        m.faceAngle.y = 0x4000
    end
    
    m.actionTimer = m.actionTimer + 1
    if m.playerIndex == 0 then
        m.pos.x = m.usedObj.oPosX - 8 * PX * sign(m.faceAngle.y)
    end

    if m.actionState == 0 and (m.usedObj == m.marioObj or m.usedObj == nil or m.usedObj.oAction ~= 0) then
        m.vel.y = PX
        if m.actionTimer % 8 == 0 then
            smb.walkCycleFrame = (smb.walkCycleFrame + 1) % 2
        end
    else
        smb.walkCycleFrame = 0
        m.vel.y = 0
    end
    m.vel.x = 0
    local landed, hitWall = smb_mario_step(m, true)
    if landed then
        smb.walkCycleFrame = 0
    end

    -- get ready to get off of vine if we reach near the top
    if m.pos.y > 5 * BLOCK then
        m.actionState = 1
    end
    return 0
end

ACT_SMB_VINE_CUTSCENE = allocate_and_hook_mario_action(act_smb_vine_cutscene, ACT_GROUP_CUTSCENE | ACT_FLAG_ON_POLE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE)

-- walk cutscene (used for flagpoles and castle end)
-- level intros (like in 1-2) use ACT_SMB_ENTER_SIDE_PIPE)
function act_smb_walk_cutscene(m)
    local smb = gSMBStates[m.playerIndex]
    smb.crouch = false

    -- return to title
    if m.playerIndex == 0 and m.actionState >= 4 and localPressStart then
        entranceType = 0
        warp_to_level(LEVEL_SMB, 1, 0)
        return 0
    end

    m.vel.y = smb_vertical_acceleration(m, smb.gravity)

    local approachSpeed = MAX_WALK_VEL
    local accel = WALK_ACCEL

    if m.playerIndex == 0 and m.actionState == 0 then
        if m.actionArg == 0 then
            set_current_music("stageClear")
        else
            set_current_music("worldClear")
        end
    end

    if m.actionState == 0 then
        m.faceAngle.y = 0x4000
        if m.vel.x < -SKID_TURN_VEL then
            accel = SKID_ACCEL
        end
    else
        approachSpeed = 0
        accel = SKID_ACCEL
        -- tick down time
        if m.playerIndex == 0 and m.actionArg == 0 and smb.timer > 0 then
            if gGlobalSyncTable.disableTimer then
                smb.timer = 0
            else
                smb.timer = smb.timer - 1
                smb.points = smb.points + 50
                play_smb_sound_stream(m.pos, "tick")
            end
        end
    end
    
    m.actionTimer = m.actionTimer + 1
    local changeTime = 8
    if math.abs(m.vel.x) >= MAX_RUN_VEL - 1 then
        changeTime = 2
    elseif math.abs(m.vel.x) >= MAX_WALK_VEL - 1 then
        changeTime = 4
    end
    if m.vel.x ~= 0 and m.actionTimer % changeTime == 0 then
        smb.walkCycleFrame = (smb.walkCycleFrame + 1) % 3
    end
    m.vel.x = smb_horizontal_acceleration(m, accel, approachSpeed)

    local landed, hitWall = smb_mario_step(m, m.actionArg ~= 0)

    -- manually move camera
    if m.actionArg ~= 0 and m.actionState < 2 then
        smb.cameraX = smb.cameraX + floor_or_ceil(math.max(m.vel.x,MAX_WALK_VEL)) / PX
        local activeScrollStop = (loaded_tile_data.scrollStop or 0) * BLOCK
        if smb.tempScrollStop ~= 0 then
            activeScrollStop = smb.tempScrollStop
        end

        if smb.cameraX < 0 then
            smb.cameraX = 0
        elseif activeScrollStop ~= 0 and smb.cameraX > (activeScrollStop // PX) + PX - NES_WIDTH then
            smb.cameraX = (activeScrollStop // PX) + PX - NES_WIDTH
            m.actionState = 2
            m.actionTimer = 0
        end
    end
    
    local np0 = gNetworkPlayers[0]
    if m.actionState == 2 and m.actionTimer >= 125 then
        -- display next part of text (not in ending)
        m.actionState = m.actionState + 1
        m.actionTimer = 0
    elseif m.actionState == 3 then
        if np0.currActNum > 28 then
            if m.actionTimer > 3 * 21 then
                -- display next part of ending text
                m.actionState = m.actionState + 1
                m.actionTimer = 0
                if m.playerIndex == 0 and np0.currActNum > 28 then
                    set_current_music("ending")
                end
            end
        elseif m.playerIndex == 0 and m.actionTimer > 12 * 21 then
            -- next level
            entranceType = 0
            warp_to_level(LEVEL_SMB, 1, np0.currActNum+1)
        end
    elseif m.actionState >= 4 and m.actionState < 7 and m.actionTimer > 60 then
        -- display next part of ending text
        m.actionState = m.actionState + 1
        m.actionTimer = 0
    end

    if not landed then
        smb.walkCycleFrame = 2
    elseif floor_or_ceil(m.vel.x) == 0 then
        smb.walkCycleFrame = -1 -- idle anim
    elseif smb.walkCycleFrame == -1 then
        smb.walkCycleFrame = 0
    end
    if m.actionArg == 0 then
        if hitWall then
            m.actionState = 1
        end
    else
        local edge = (loaded_tile_data.scrollStop + 1) * BLOCK
        if m.actionState == 0 and m.pos.x >= edge - 9 * BLOCK then
            m.actionState = 1
        elseif m.actionState == 1 and m.pos.x >= edge - 8 * BLOCK then
            m.pos.x = edge - 8 * BLOCK
        end
    end

    return 0
end
ACT_SMB_WALK_CUTSCENE = allocate_and_hook_mario_action(act_smb_walk_cutscene, ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE | ACT_FLAG_MOVING)

-- action used when falling out of coin heaven
function act_smb_bonus_transition(m)
    m.actionTimer = m.actionTimer + 1

    -- do the warp
    if m.playerIndex == 0 and m.actionTimer >= 90 then
        local np = gNetworkPlayers[0]
        local area = respawnArea
        entranceType = 1
        doBonusExit = true
        warp_to_level(np.currLevelNum, area, np.currActNum)
    end
end

ACT_SMB_BONUS_TRANSITION = allocate_and_hook_mario_action(act_smb_bonus_transition, ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE)

-- hit axe action
function act_smb_hit_axe(m)
    if m.playerIndex ~= 0 then return 0 end

    disable_time_stop()
    timeStopEnabled = false

    m.actionTimer = m.actionTimer + 1
    if m.actionState == 0 then
        -- bridge break sequence
        if m.actionTimer < 14 then return 0 end

        if m.actionTimer == 14 then -- remove chain
            local sx = truncate(m.pos.x // BLOCK, 16)
            add_to_2d_table(loaded_tile_data, sx + 12, 5, 0)
            play_smb_sound_stream(m.pos, "break")
        elseif m.actionTimer % 4 == 2 then -- remove bridge piece
            local sx = truncate(m.pos.x // BLOCK, 16)
            add_to_2d_table(loaded_tile_data, sx + 12 - (m.actionTimer - 18) // 4, 4, TILE_INVISIBLE_BLOCK)
            play_smb_sound_stream(m.pos, "break")
            if 12 - (m.actionTimer - 18) // 4 <= 0 then
                m.actionState = 1
            end
        end
    else
        -- do nothing until the bowser cutscene object is gone
        local bowser = obj_get_first_with_behavior_id(id_bhvSmbBowserCutscene)
        if not bowser then
            set_mario_action(m, ACT_SMB_WALK_CUTSCENE, 1)
        end
    end

    return 0
end
ACT_SMB_HIT_AXE = allocate_and_hook_mario_action(act_smb_hit_axe, ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE | ACT_FLAG_MOVING)

-- title screen action
function act_smb_title_screen(m)
    if m.playerIndex ~= 0 or dontRenderTime ~= 0 then return 0 end
    m.actionTimer = m.actionTimer + 1
    if m.actionTimer < 3 then return 0 end

    if (not localOptionTable.debugHalfSpeed) then
        if m.actionState == 0 then
            m.actionState = 1
        else
            m.actionState = 0
            return 0
        end
    end
    
    if m.controller.buttonPressed & B_BUTTON ~= 0 and gGlobalSyncTable.allowWorldSelect then
        startWorld = (startWorld % 8) + 1
    elseif m.controller.buttonPressed & R_TRIG ~= 0 then
        djui_open_pause_menu()
    elseif m.controller.buttonPressed & U_JPAD ~= 0 or m.controller.buttonPressed & D_JPAD ~= 0 then
        local totalOptions = 3
        if gGlobalSyncTable.topWorld <= 1 then totalOptions = totalOptions - 1 end

        local change = 1
        if m.controller.buttonPressed & U_JPAD ~= 0 then
            change = -1
        end
        smbMenuOption = (smbMenuOption + change) % totalOptions
    elseif localPressStart or m.controller.buttonPressed & A_BUTTON ~= 0 then
        local testOption = smbMenuOption
        if gGlobalSyncTable.topWorld <= 1 then testOption = testOption + 1 end
        if testOption == 0 then
            -- start game from top world
            warp_to_level(LEVEL_SMB, 1, gGlobalSyncTable.topWorld*4-3)
        elseif testOption == 1 then
            -- start game from selected world
            warp_to_level(LEVEL_SMB, 1, startWorld*4-3)
        elseif testOption == 2 then
            -- return to start level
            warp_to_level(gLevelValues.entryLevel, 1, 0)
        end
    end
    return 0
end
ACT_SMB_TITLE_SCREEN = allocate_and_hook_mario_action(act_smb_title_screen, ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE | ACT_FLAG_STATIONARY)

-- status screen action (before starting a level)
function act_smb_loading(m)
    if m.playerIndex ~= 0 then return 0 end

    if m.invincTimer ~= 0 then m.invincTimer = m.invincTimer + 1 end

    m.actionTimer = m.actionTimer + 1
    if m.actionArg == 1 then
        set_current_music("gameover")
        if timeUpDeath and m.actionTimer > 7 * 21 then
            timeUpDeath = false
            m.actionTimer = 0
        elseif m.actionTimer > 12 * 21 then
            warp_to_level(LEVEL_SMB, 1, 0)
        end
    elseif timeUpDeath then
        if m.actionTimer > 7 * 21 then
            timeUpDeath = false
            m.actionTimer = 0
        end
    elseif m.actionTimer > 7 * 21 and checkFirstHereStatus ~= -1 then
        local newAction = m.prevAction
        if newAction == m.action then
            newAction = ACT_SMB_IDLE
        end
        set_music_to_theme()
        set_mario_action(m, newAction, 0)
    end
    return 0
end
ACT_SMB_LOADING = allocate_and_hook_mario_action(act_smb_loading, ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE | ACT_FLAG_INVULNERABLE | ACT_FLAG_STATIONARY)

-- force SMB actions
function before_set_action(m, action)
    if gNetworkPlayers[0].currLevelNum ~= LEVEL_SMB then return end
    if action == ACT_IDLE then
        return ACT_SMB_IDLE
    elseif action == ACT_JUMP then
        return ACT_SMB_JUMP
    elseif action == ACT_FREEFALL then
        return ACT_SMB_FALL
    end
    
    -- switch to water actions if this is a water stage
    if loaded_tile_data and loaded_tile_data.theme == 3 then
        if action == ACT_SMB_IDLE then
            return ACT_SMB_WATER_IDLE
        elseif action == ACT_SMB_WALKING then
            return ACT_SMB_WATER_WALKING
        elseif action == ACT_SMB_JUMP or action == ACT_SMB_FALL then
            return ACT_SMB_SWIMMING
        end
    end
end
hook_event(HOOK_BEFORE_SET_MARIO_ACTION, before_set_action)

function on_set_action(m)
    if gNetworkPlayers[0].currLevelNum ~= LEVEL_SMB then return end

    local smb = gSMBStates[m.playerIndex]
    if m.action == ACT_SMB_JUMP then
        -- set initial y speed and gravity
        m.vel.y = JUMP_VEL
        smb.initialJumpXVel = m.vel.x -- used as "initial jump velocity"
        if math.abs(m.vel.x) < 0x10 then
            smb.gravity = FALL_GRAVITY_SLOW
            smb.holdAGravity = HOLD_A_GRAVITY_SLOW
        elseif math.abs(m.vel.x) <= 0x25 then
            smb.gravity = FALL_GRAVITY
            smb.holdAGravity = HOLD_A_GRAVITY
        else
            m.vel.y = JUMP_VEL_FAST
            smb.gravity = FALL_GRAVITY_FAST
            smb.holdAGravity = HOLD_A_GRAVITY_FAST
        end
    elseif m.action == ACT_SMB_SWIMMING and m.actionArg == 1 then
        -- set initial y speed and gravity
        smb.initialJumpXVel = m.vel.x -- used as "initial jump velocity"
        smb.walkCycleFrame = 0

        -- set gravity and y vel based on what "zone" we are in
        if m.pos.y > NES_HEIGHT * PX - 4 * BLOCK then
            smb.gravity = GRAVITY_WATER_SURFACE
            smb.holdAGravity = GRAVITY_WATER_SURFACE
            if m.pos.y <= NES_HEIGHT * PX - 3.5 * BLOCK then
                m.vel.y = SWIM_VEL
            end
        elseif smb.inWhirlpool then
            smb.gravity = FALL_GRAVITY_WATER_WHIRLPOOL
            smb.holdAGravity = HOLD_A_GRAVITY_WATER_WHIRLPOOL
            m.vel.y = SWIM_VEL_WHIRLPOOL
        else
            smb.gravity = FALL_GRAVITY_WATER
            smb.holdAGravity = HOLD_A_GRAVITY_WATER
            m.vel.y = SWIM_VEL
        end
    end
end
hook_event(HOOK_ON_SET_MARIO_ACTION, on_set_action)