-- audio handling

SMB_SOUNDS = {
    jumpSmall = audio_sample_load("smb_jump-small.mp3"),
    jumpBig = audio_sample_load("smb_jump-super.mp3"),
    bump = audio_sample_load("smb_bump.mp3"),
    stomp = audio_sample_load("smb_stomp.mp3"),
    kick = audio_sample_load("smb_kick.mp3"),
    pipe = audio_sample_load("smb_pipe.mp3"),
    deathShort = audio_sample_load("smb_mariodie-short.mp3"),
    coin = audio_sample_load("smb_coin.mp3"),
    spawnItem = audio_sample_load("smb_powerup_appears.mp3"),
    pause = audio_sample_load("smb_pause.mp3"),
    powerup = audio_sample_load("smb_powerup.mp3"),
    boom = audio_sample_load("smb_fireworks.mp3"),
    fireball = audio_sample_load("smb_fireball.mp3"),
    warning = audio_sample_load("smb_warning.mp3"),
    flagpole = audio_sample_load("smb_flagpole.mp3"),
    bowserDeath = audio_sample_load("smb_bowserfalls.mp3"),
    bowserFlame = audio_sample_load("smb_bowserfire.mp3"),
    vine = audio_sample_load("smb_vine.mp3"),
    tick = audio_sample_load("smb_tick.mp3"),
    ["break"] = audio_sample_load("smb_breakblock.mp3"),
    ["1up"] = audio_sample_load("smb_1-up.mp3"),
}
SMB_SOUNDS.hurt = SMB_SOUNDS.pipe
SMB_SOUNDS.swim = SMB_SOUNDS.stomp

-- plays the streamed sound if the MARIO_ACTION_SOUND_PLAYED flag is 0, and sets it to 1
function play_mario_action_sound_stream(m, soundName)
    if not m then return end
    if m.flags & MARIO_ACTION_SOUND_PLAYED == 0 then
        if gSMBStates[0].paused or gMarioStates[0].action == ACT_SMB_LOADING then
            -- nothing
        elseif SMB_SOUNDS[soundName] then
            audio_sample_play(SMB_SOUNDS[soundName], m.pos, 0.5)
        else
            log_to_console("No audio file: "..soundName)
        end
        m.flags = m.flags | MARIO_ACTION_SOUND_PLAYED
    end
end

-- plays the streamed sound
function play_smb_sound_stream(pos, soundName, playWhenPaused)
    if not pos then return end
    if (gSMBStates[0].paused or gMarioStates[0].action == ACT_SMB_LOADING) and not playWhenPaused then return end
    if SMB_SOUNDS[soundName] then
        audio_sample_play(SMB_SOUNDS[soundName], pos, 0.5)
    else
        log_to_console("No audio file: "..soundName)
    end
end

-- plays the streamed sound relative to this object
-- if deathSound is true, it sets the death sound flag and doesn't play the sound if it's already set
function play_smb_sound_object(o, soundName, deathSound, playWhenPaused)
    local pos = {x = o.oPosX, y = o.oPosY, z = o.oPosZ}
    if deathSound then
        if o.oDeathSound ~= 0 then return end
        o.oDeathSound = 1
    end
    play_smb_sound_stream(pos, soundName, playWhenPaused)
end

--- music ---
SMB_MUSIC = {
    ground = {music = audio_stream_load("ground.mp3"), loopStart = 2.14, loopEnd = 81.4, speedup = "groundHurry"},
    underground = {music = audio_stream_load("underground.mp3"), loopStart = 0, loopEnd = 12.6, speedup = "undergroundHurry"},
    castle = {music = audio_stream_load("castle.mp3"), loopStart = 0, loopEnd = 8.1, speedup = "castleHurry"},
    underwater = {music = audio_stream_load("underwater.mp3"), loopStart = 0, loopEnd = 25.5, speedup = "underwaterHurry"},
    star = {music = audio_stream_load("star.mp3"), loopStart = 0, loopEnd = 3.1869, speedup = "starHurry"},
    groundHurry = {music = audio_stream_load("groundHurry.mp3"), loopStart = 0, loopEnd = -1},
    undergroundHurry = {music = audio_stream_load("undergroundHurry.mp3"), loopStart = 0, loopEnd = -1},
    castleHurry = {music = audio_stream_load("castleHurry.mp3"), loopStart = 0, loopEnd = -1},
    underwaterHurry = {music = audio_stream_load("underwaterHurry.mp3"), loopStart = 0, loopEnd = -1},
    starHurry = {music = audio_stream_load("starHurry.mp3"), loopStart = 0, loopEnd = -1},
    stageClear = {music = audio_stream_load("smb_stage_clear.mp3"), volume = 0.5},
    worldClear = {music = audio_stream_load("smb_world_clear.mp3"), volume = 0.5},
    death = {music = audio_stream_load("smb_mariodie.mp3"), volume = 0.5},
    gameover = {music = audio_stream_load("smb_gameover.mp3"), volume = 0.5},
    ending = {music = audio_stream_load("ending.mp3"), loopStart = 0, loopEnd = 6.4},
}
SMB_MUSIC.pipeIntro = {music = SMB_MUSIC.ground.music, stop = 2.15}

if audio_stream_set_loop_points then
    for name,data in pairs(SMB_MUSIC) do
        if data.loopEnd then
            audio_stream_set_looping(data.music, true)
            if data.loopEnd ~= -1 then
                local sampleRate = 44100 -- 44.1 kHz
                local sampleStart = data.loopStart * sampleRate
                local sampleEnd = data.loopEnd * sampleRate
                audio_stream_set_loop_points(data.music, sampleStart, sampleEnd)
            end
        end
    end
end

local currentMusic = ""
local newMusic = ""
local musicPausedPos = -1
playWarningTimer = 0
function update_music()
    local musicToPlay = newMusic or ""
    local starActive = false
    if gNetworkPlayers[0].currLevelNum ~= LEVEL_SMB then
        musicToPlay = ""
        newMusic = ""
    elseif gSMBStates[0].starTimer and gSMBStates[0].starTimer > 3 * 21 and gMarioStates[0].action & ACT_GROUP_MASK ~= ACT_GROUP_CUTSCENE then
        -- auto-play star music
        musicToPlay = "star"
        starActive = true
        if gSMBStates[0].timer < 100 then
            musicToPlay = "starHurry"
        end
    end
    
    if playWarningTimer ~= 0 then -- stop music while the warning sfx is playing
        playWarningTimer = playWarningTimer - 1

        -- if we're in a cutscene, ignore warning
        if gMarioStates[0].action & ACT_GROUP_MASK == ACT_GROUP_CUTSCENE then
            audio_sample_stop(SMB_SOUNDS.warning)
            playWarningTimer = 0
        end

        if currentMusic == "" then return end

        local musicData = SMB_MUSIC[currentMusic]
        local music = musicData.music
        audio_stream_pause(music)
        return
    elseif gSMBStates[0].paused then -- pause music
        if currentMusic == "" then return end

        local musicData = SMB_MUSIC[currentMusic]
        local music = musicData.music
        musicPausedPos = audio_stream_get_position(music)
        audio_stream_pause(music)
        return
    elseif musicPausedPos ~= -1 then -- unpause music
        if currentMusic == "" then return end

        local musicData = SMB_MUSIC[currentMusic]
        local music = musicData.music
        audio_stream_play(music, false, musicData.volume or 1)
        audio_stream_set_position(music, musicPausedPos)
        musicPausedPos = -1
    end

    if currentMusic ~= musicToPlay then
        if currentMusic ~= "" then
            local musicData = SMB_MUSIC[currentMusic]
            local music = musicData.music
            audio_stream_stop(music)
        end
        currentMusic = musicToPlay
        if currentMusic == "" then return end

        local musicData = SMB_MUSIC[currentMusic]
        if musicData == nil then
            log_to_console("No music found: "..currentMusic)
            currentMusic = ""
            if not starActive then
                newMusic = ""
            end
            return
        end
        local music = musicData.music

        audio_stream_play(music, false, musicData.volume or 1)
        return
    end

    if currentMusic == "" then return end

    local musicData = SMB_MUSIC[currentMusic]
    local music = musicData.music

    -- play music faster if time is low
    if (not starActive) and musicData.speedup and gSMBStates[0].timer < 100 then
        newMusic = musicData.speedup
    end

    -- do looping/stopping for pre 1.3 versions... and also for the ground track because it doesn't work fsr
    local currPosition = audio_stream_get_position(music)
    if (not audio_stream_set_loop_points) or currentMusic == "ground" then
        local loopEnd = musicData.loopEnd or 0
        if loopEnd ~= 0 then
            audio_stream_set_looping(music, true)
            if loopEnd ~= -1 and currPosition >= loopEnd then
                local minus = loopEnd - (musicData.loopStart or 0)
                audio_stream_set_position(music, currPosition - minus)
            end
        end
    end

    if musicData.stop and currPosition >= musicData.stop then
        audio_stream_stop(music)
    end
end
hook_event(HOOK_UPDATE, update_music)

-- sets the currently played music; set an empty string or nil to play nothing
function set_current_music(trackName)
    newMusic = trackName
end

-- sets the background music to whatever is appropriate for this theme
function set_music_to_theme()
    if loaded_tile_data.pipeCutscene then
        newMusic = "pipeIntro"
    elseif loaded_tile_data.bonusArea then
        newMusic = "star"
    elseif loaded_tile_data.theme == 0 then
        newMusic = "ground"
    elseif loaded_tile_data.theme == 1 then
        newMusic = "underground"
    elseif loaded_tile_data.theme == 2 then
        newMusic = "castle"
    elseif loaded_tile_data.theme == 3 then
        newMusic = "underwater"
    else
        newMusic = "ground" -- default
    end

    -- restart music if it is the same
    if currentMusic == newMusic and currentMusic ~= "" then
        local musicData = SMB_MUSIC[currentMusic]
        local music = musicData.music
        audio_stream_play(music, true, musicData.volume or 1)
    end
end

function loop_test_command()
    if currentMusic ~= "" then
        local musicData = SMB_MUSIC[currentMusic]
        local music = musicData.music
        if musicData.loopEnd then
            audio_stream_set_position(music, musicData.loopEnd - 3)
        end
    end
    return true
end

function timer_test_command()
    gSMBStates[0].timer = 99
    play_smb_sound_stream(gMarioStates[0].pos, "warning", true)
    playWarningTimer = 3 * 30
    return true
end

function music_test_command(msg)
    newMusic = msg
    return true
end

if _G.cheatsApi then
    hook_chat_command("looptest", "- Go near the loop point of this song", loop_test_command)
    hook_chat_command("timertest", "- Trigger \"hurry\" theme", timer_test_command)
    hook_chat_command("musictest", "[NAME] - Play the specified track", music_test_command)
end