--- packet handling (mainly for level loading and block interaction) ---

-- process a block interaction
function on_packet_block_interact(data, self)
    local tempTile = TILE_INVISIBLE_BLOCK
    local replaceTile = TILE_BRICK
    local sound = ""
    local doSave = true
    local coordX = (data.x + 0.5) * BLOCK
    local coordY = data.y * BLOCK
    local inCurrentArea = (self or gNetworkPlayers[0].currAreaIndex == data.area)

    if data.tile == TILE_COIN then
        tempTile = 0
        replaceTile = 0
        if inCurrentArea and data.hitFromBelow then
            do_coin_particle(data.x, data.y, self)
        elseif self then
            local smb = gSMBStates[0]
            smb.points = smb.points + 200
            gGlobalSyncTable.coins = gGlobalSyncTable.coins + 1
        end
    elseif data.tile == TILE_QUESTION_COIN or data.tile == TILE_HIDDEN_COIN then
        if data.tile ~= TILE_HIDDEN_COIN then
            replaceTile = TILE_HIT_BLOCK
        else
            replaceTile = TILE_HIT_BRICK
        end
        if inCurrentArea then
            do_coin_particle(data.x, data.y + 1, self)
        end
    elseif data.tile == TILE_BRICK_1UP or data.tile == TILE_HIDDEN_1UP then
        replaceTile = TILE_HIT_BRICK
        doSave = false
        sound = "spawnItem"
        if self then
            spawn_object_no_rotate(id_bhvSmbMushroom, E_MODEL_NONE, coordX, coordY, 0, true, function(o)
                o.oBehParams = 1
            end)
        end
    elseif data.tile == TILE_QUESTION_ITEM or data.tile == TILE_BRICK_ITEM then
        if data.tile ~= TILE_BRICK_ITEM then
            replaceTile = TILE_HIT_BLOCK
        else
            replaceTile = TILE_HIT_BRICK
        end
        doSave = false
        sound = "spawnItem"
        if self then
            local smb0 = gSMBStates[0]
            if smb0.powerup == 0 then
                spawn_object_no_rotate(id_bhvSmbMushroom, E_MODEL_NONE, coordX, coordY, 0, true, nil)
            else
                spawn_object_no_rotate(id_bhvSmbFireFlower, E_MODEL_NONE, coordX, coordY, 0, true, nil)
            end
        end
    elseif data.tile == TILE_BRICK_STAR then
        replaceTile = TILE_HIT_BRICK
        doSave = false
        sound = "spawnItem"
        if self then
            spawn_object_no_rotate(id_bhvSmbSuperStar, E_MODEL_NONE, coordX, coordY, 0, true, nil)
        end
    elseif data.tile == TILE_BRICK_COINS then
        if data.coinBrickTimer == nil then
            data.coinBrickTimer = 11 * 21 -- I'm not recreating framerules sorry
        end

        if data.coinBrickTimer ~= 0 then
            replaceTile = TILE_BRICK_COINS
        else
            replaceTile = TILE_HIT_BRICK
        end
        coin_brick_timer[data.x] = data.coinBrickTimer
        if inCurrentArea then
            do_coin_particle(data.x, data.y + 1, self)
        end
    elseif data.tile == TILE_BRICK_VINE then
        replaceTile = TILE_HIT_BRICK
        sound = "" -- vine object plays the sound
        if self then
            spawn_object_no_rotate(id_bhvSmbVine, E_MODEL_NONE, coordX, coordY, 0, true, nil)
        end
    elseif data.big then -- break brick
        sound = "break"
        tempTile = 0
        replaceTile = 0

        -- brick particles
        if inCurrentArea then
            for i=0,3 do
                local coordX2 = coordX + (i % 2 - 0.5) * BLOCK * 0.5
                local coordY2 = coordY + (i // 2) * BLOCK * 0.5
                spawn_object_no_rotate(id_bhvSmbBrickParticle, E_MODEL_NONE, coordX2, coordY2, 0, false, function(o)
                    o.oVelX = (i % 2 - 0.5) * 2 * PX
                    if i // 2 == 0 then
                        o.oVelY = 50
                    else
                        o.oVelY = 80
                    end
                end)
            end
        end

        if self then
            local smb = gSMBStates[0]
            smb.points = smb.points + 50
        end
    else
        doSave = false -- don't save just bouncing a brick
    end
    
    -- format weirdly to fit in packet
    local height = 13
    if self then
        if not interacted_blocks_local[data.area] then
            interacted_blocks_local[data.area] = {}
        end
        interacted_blocks_local[data.area][data.y + data.x * height] = replaceTile
    end

    if doSave then
        if not interacted_blocks[data.area] then
            interacted_blocks[data.area] = {}
        end
        interacted_blocks[data.area][data.y + data.x * height] = replaceTile
    elseif data.tile == TILE_HIDDEN_1UP then -- for hidden 1ups, replace with a brick for remote players
        local packetTile = TILE_BRICK_1UP
        if not interacted_blocks[data.area] then
            interacted_blocks[data.area] = {}
        end
        interacted_blocks[data.area][data.y + data.x * height] = packetTile
        if not self then
            replaceTile = packetTile
        end
    elseif not self then -- for remote players, don't change what block this is (used for powerup blocks)
        if interacted_blocks_local[data.area] and interacted_blocks_local[data.area][data.y + data.x * height] then
            replaceTile = interacted_blocks_local[data.area][data.y + data.x * height]
        else
            replaceTile = data.tile
        end
    end

    if not inCurrentArea then return end

    -- bump remote players
    if (not self) and data.hitFromBelow and data.from and data.tile ~= TILE_COIN then
        local index = network_local_index_from_global(data.from)
        if index ~= 255 then
            local m = gMarioStates[index]
            play_smb_sound_stream(m.pos, "bump")
            if data.tile == TILE_BRICK then
                m.vel.y = -0x20
            else
                m.vel.y = 0
            end
        end
    end

    if sound ~= "" then
        play_smb_sound_stream({x = coordX, y = coordY, z = 0}, sound)
    end
    add_to_2d_table(loaded_tile_data, data.x, data.y, tempTile)
    -- spawn bounce block object
    if replaceTile ~= 0 then
        spawn_object_no_rotate(id_bhvSmbBounceBlock, E_MODEL_NONE, coordX, coordY, 0, false, function(o)
            o.oBehParams = replaceTile
        end)
    end

    -- object interaction
    if data.hitFromBelow and data.tile ~= TILE_COIN then
        local index = network_local_index_from_global(data.from)
        if index ~= 255 then
            local m = gMarioStates[index]
            hit_from_below_interactions(m, data.x, data.y)
        end
    end

    -- collect coins above this tile
    if data.tile ~= TILE_COIN and get_tile_at_block(data.x, data.y + 1) == TILE_COIN then
        on_packet_block_interact({
            id = PACKET_BLOCK_INTERACT,
            x = data.x,
            y = data.y + 1,
            tile = TILE_COIN,
            hitFromBelow = true,
            area = data.area,
            from = data.from,
        }, self)
    end
end

-- either send or receive block interaction data
function on_packet_blocks_request(data, self)
    if data.load then
        if not self then
            interacted_blocks[data.area] = data
        end

        local height = 13
        for coord,tile in pairs(data) do
            if tonumber(coord) then
                local x = coord // height
                local y = coord % height
                add_to_2d_table(loaded_tile_data, x, y, tile)
                if tile == TILE_BRICK_COINS then
                    coin_brick_timer[x] = 0
                end
            end
        end
        return
    end
    local fromLocal = network_local_index_from_global(data.from)
    if fromLocal == 255 then return end
    if not interacted_blocks[data.area] then return end
    interacted_blocks[data.area].id = PACKET_BLOCKS_REQUEST
    interacted_blocks[data.area].load = true
    interacted_blocks[data.area].area = data.area
    network_send_to(fromLocal, true, interacted_blocks[data.area])
end

function on_packet_check_if_first(data, self)
    if data.returnStatus then
        local np = gNetworkPlayers[0]
        -- if we were waiting for this packet, run object spawn now
        if np.currAreaSyncValid and np.currLevelSyncValid and checkFirstHereStatus == -1 then
            initial_object_load(data.returnStatus == 1)
        end
        checkFirstHereStatus = data.returnStatus
        return
    end

    local np = network_player_from_global_index(data.from)
    if not np then return end

    -- check if someone else already sent a packet
    local status = 1
    for i,data2 in ipairs(recentlyLoaded) do
        if data2.levelNum == data.levelNum and data2.area == data.area then
            status = 0
            --log_to_console(tostring(data.from).. " is not first in "..tostring(data.area).." "..tostring(data.levelNum).." because of recentlyLoaded")
            break
        end
    end
    
    -- check if no one else is present in the stage
    if status ~= 0 then
        for i=0,MAX_PLAYERS-1 do
            local np2 = gNetworkPlayers[i]
            if i ~= np.localIndex and np2.connected and np2.currAreaSyncValid and np2.currLevelSyncValid and np2.currLevelNum == LEVEL_SMB and np2.currActNum == data.levelNum and np2.currAreaIndex == data.area then
                status = 0
                --log_to_console(tostring(data.from).. " is not first in "..tostring(data.area).." "..tostring(data.levelNum).." because of other player "..tostring(np2.globalIndex))
                break
            end
        end
    end

    if status ~= 0 then
        --log_to_console(tostring(data.from) .. " was first in "..tostring(data.area).." "..tostring(data.levelNum).."! Sending status...")
        table.insert(recentlyLoaded, {
            levelNum = data.levelNum,
            area = data.area,
            index = np.localIndex,
        })
    end

    -- if we're checking this packet ourselves, just set our status
    if self then
        checkFirstHereStatus = status
        return
    end

    -- otherwise, send packet back
    network_send_to(np.localIndex, true, {
        id = PACKET_CHECK_IF_FIRST,
        returnStatus = status,
    })
end

function on_packet_emergency_load()
    initial_object_load(true, true)
end

function on_packet_game_reset()
    local np = gNetworkPlayers[0]
    if np.currLevelNum == LEVEL_SMB and np.currActNum ~= 0 then
        warp_to_level(LEVEL_SMB, 1, 0)
    end
end

interacted_blocks = {}
interacted_blocks_local = {}
coin_brick_timer = {}
PACKET_BLOCK_INTERACT = 0
PACKET_BLOCKS_REQUEST = 1
PACKET_CHECK_IF_FIRST = 2
PACKET_EMERGENCY_LOAD = 3
PACKET_GAME_RESET = 4
local sPacketTable = {
    [PACKET_BLOCK_INTERACT] = on_packet_block_interact,
    [PACKET_BLOCKS_REQUEST] = on_packet_blocks_request,
    [PACKET_CHECK_IF_FIRST] = on_packet_check_if_first,
    [PACKET_EMERGENCY_LOAD] = on_packet_emergency_load,
    [PACKET_GAME_RESET] = on_packet_game_reset,
}

function on_packet_receive(data)
    if sPacketTable[data.id] then
        sPacketTable[data.id](data, false)
    end
end
hook_event(HOOK_ON_PACKET_RECEIVE, on_packet_receive)

-- send packet, and also run this packet's code ourselves
function network_send_include_self(reliable, data)
    network_send(reliable, data)
    if sPacketTable[data.id] then
        sPacketTable[data.id](data, true)
    end
end

-- sends packet to every player in our current level and act, and also runs for ourselves
function network_send_to_all_in_act(reliable, data)
    for i=1,MAX_PLAYERS-1 do
        local np = gNetworkPlayers[i]
        if np.connected and (np.currLevelNum == gNetworkPlayers[0].currLevelNum and np.currActNum == gNetworkPlayers[0].currActNum) and np.currLevelSyncValid then
            network_send_to(i, reliable, data)
        end
    end
    if sPacketTable[data.id] then
        sPacketTable[data.id](data, true)
    end
end

-- send packet to the host... unless we ARE the host, in which case just run the packet ourselves
function network_send_to_host(reliable, data)
    if network_is_server() then
        if sPacketTable[data.id] then
            sPacketTable[data.id](data, true)
        end
    else
        network_send_to(1, reliable, data)
    end
end